import asyncio
import csv
import logging
import os
import re
import time
from pathlib import Path
from pprint import pformat, pprint
from typing import Any, List, Optional

from .api import BunkrrAPI
from .cli import cli

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.WARNING)


class BunkrrUploader:
    def __init__(
        self,
        token: str,
        max_connections: int = 1,
        retries: int = 1,
        options: Optional[dict[str, Any]] = None,
    ):
        if options is None:
            options = {}
        self.options = options
        self.api = BunkrrAPI(token, max_connections, retries, options)
        self.temporary_files = []

    async def init(self):
        raw_req = await self.api.get_check()
        logger.debug(pformat(raw_req))
        max_file_size = raw_req.get("maxSize", "0B")
        max_chunk_size = raw_req.get("chunkSize", {}).get("max", "0B")
        default_chunk_size = raw_req.get("chunkSize", {}).get("default", "0B")
        self.api.file_blacklist.extend(raw_req.get("stripTags", {}).get("blacklistExtensions", []))

        # Choose a chunk size, default or max
        chunk_size = default_chunk_size
        # chunk_size = '1MB'

        if max_file_size == "0B" or chunk_size == "0B":
            raise Exception("Invalid max file size or chunk size")

        # TODO: check if either one is 0 and abort

        units_to_calc = [max_file_size, chunk_size]
        units_calculated = []

        for unit in units_to_calc:
            size_str = unit.lower()
            unit_multiplier = {"b": 1, "kb": 1024, "mb": 1024**2, "gb": 1024**3, "tb": 1024**4}
            match = re.match(r"^(\d+)([a-z]+)$", size_str)

            if match:
                value, unit = match.groups()
                bytes_size = int(value) * unit_multiplier.get(unit, 1)
                units_calculated.append(bytes_size)
            else:
                raise ValueError("Invalid input format")

        self.api.max_file_size = units_calculated[0]
        self.api.chunk_size = units_calculated[1]

    def prepare_file_for_upload(self, file: Path) -> List[Path]:
        file_size = os.stat(file).st_size

        # TODO: Truncate the file name if it is too long
        file_name = (file.name[:240] + "..") if len(file.name) > 240 else file.name

        if file.suffix in self.api.file_blacklist:
            logger.error(f"File {file} has blacklisted extension {file.suffix}")
            return []

        if file_size > self.api.max_file_size:
            # TODO: Create temporary file archive
            logger.error(f"File {file} is bigger than max file size {self.api.max_file_size}")
            return []

        return [file]

    async def upload_files(self, path: Path, folder: Optional[str] = None) -> None:
        if path.is_file():
            paths = [path]
        else:
            paths = [x for x in path.iterdir() if x.is_file()]
            if folder is None:
                folder = path.name

        # The server may not accept certain file types and those over a certain size so we need to create temporary files
        filtered_paths = []
        for file_path in paths:
            filtered_paths.extend(self.prepare_file_for_upload(file_path))

        if len(paths) == 0:
            print("No file paths left to upload")
            return

        # TODO: Delete the extra created files after upload
        self.temporary_files = [x for x in filtered_paths if x not in paths]

        folder_id = None
        if folder:
            existing_folders = await self.api.get_albums()
            existing_folder = next((x for x in existing_folders["albums"] if x["name"] == folder), None)
            if existing_folder:
                folder_id = str(existing_folder["id"])
            else:
                created_folder = await self.api.create_album(folder, folder)
                folder_id = str(created_folder["id"])

        if paths:
            responses = await self.api.upload_files(filtered_paths, folder_id)
            pprint(responses)

            if self.options.get("save") is True and responses:
                file_name = f"bunkrr_upload_{int(time.time())}.csv"
                with open(file_name, "w", newline="") as csvfile:
                    logger.info(f"Saving uploaded files to {file_name}")
                    field_names = [
                        "name",
                        "url",
                        "fileName",
                        "albumid",
                        "uploadSuccess",
                        "filePath",
                        "filePathMD5",
                        "fileNameMD5",
                    ]
                    csv_writer = csv.DictWriter(csvfile, dialect="excel", fieldnames=field_names)
                    csv_writer.writeheader()
                    for row in responses:
                        csv_writer.writerow(row["files"][0])

            else:
                pprint(responses)


async def async_main() -> None:
    args = cli()
    logger.debug(args)

    options = {
        "save": args.save,
        "chunk_retries": args.chunk_retries
    }

    bunkrr_client = BunkrrUploader(args.token, max_connections=args.connections, retries=args.retries, options=options)
    try:
        await bunkrr_client.init()
        if args.dry_run:
            print("Dry run only, uploading skipped")
        else:
            await bunkrr_client.upload_files(args.file, folder=args.folder)
    finally:
        if not bunkrr_client.api.session.closed:
            await bunkrr_client.api.session.close()
        for server_session in bunkrr_client.api.server_sessions.values():
            if not server_session.closed:
                server_session.close()


def main():
    asyncio.run(async_main())



from typing import List, TypedDict


class ChunkSize(TypedDict):
    max: str
    default: str
    timeout: int


class FileIdentifierLength(TypedDict):
    min: int
    max: int
    default: int
    force: bool


class StripTags(TypedDict):
    default: bool
    video: bool
    force: bool
    blacklistExtensions: List[str]


class CheckResponse(TypedDict):
    private: bool
    enableUserAccounts: bool
    maxSize: str
    chunkSize: ChunkSize
    fileIdentifierLength: FileIdentifierLength
    stripTags: StripTags
    temporaryUploadAges: List[int]
    defaultTemporaryUploadAge: int


class NodeResponse(TypedDict):
    success: bool
    url: str


class Permissions(TypedDict):
    user: bool
    vip: bool
    vvip: bool
    moderator: bool
    admin: bool
    superadmin: bool


class VerifyTokenResponse(TypedDict):
    success: bool
    username: str
    permissions: Permissions
    group: str
    retentionPeriods: List[int]
    defaultRetentionPeriod: int


class File(TypedDict):
    name: str
    url: str
    # The are hacked on and not officially there from the API
    original: str
    albumid: str
    filePathMD5: str
    fileNameMD5: str
    uploadSuccess: str


# This is the same for finish chunks
class UploadResponse(TypedDict):
    success: bool
    files: List[File]


class AlbumItemResponse(TypedDict):
    id: int
    name: str
    identifier: str


class AlbumsResponse(TypedDict):
    success: bool
    albums: List[AlbumItemResponse]
    count: int


class CreateAlbumResponse(TypedDict):
    success: bool
    id: int



import asyncio
import hashlib
import logging
import mimetypes
import os
import uuid
from pathlib import Path
from pprint import pformat, pprint
from typing import Any, BinaryIO, Optional, Union

import aiohttp
from tqdm.asyncio import tqdm_asyncio

from .types import (
    AlbumsResponse,
    CheckResponse,
    CreateAlbumResponse,
    NodeResponse,
    UploadResponse,
    VerifyTokenResponse,
)
from .util import ProgressFileReader, TqdmUpTo

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class BunkrrAPI:
    def __init__(
        self,
        token: str,
        max_connections: int = 2,
        retries: int = 2,
        options: Optional[dict[str, Any]] = None,
    ):
        if options is None:
            options = {}

        self.token = token
        self.url = "https://app.bunkrr.su/"
        self.download_url_base = "https://bunkrr.ru/d/"

        # These all need to be initialized later on before the API is used
        self.max_file_size = None
        self.chunk_size = None
        self.file_blacklist = []

        self.options = options

        self.session_headers = {
            "Accept": "application/json",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
            "token": self.token,
        }
        self.session = aiohttp.ClientSession("https://app.bunkrr.su", headers=self.session_headers)

        self.server_sessions = {}
        self.created_folders = {}
        self.sem = asyncio.Semaphore(max_connections)
        self.retries = retries
        self.max_chunk_retries = options.get('chunk_retries') or 1

    async def get_check(self) -> CheckResponse:
        async with self.session.get("/api/check") as resp:
            response = await resp.json()
            return response

    async def get_node(self) -> NodeResponse:
        async with self.session.get("/api/node") as resp:
            response = await resp.json()
            return response

    async def verify_token(self) -> VerifyTokenResponse:
        data = {"token": self.token}
        async with self.session.post("/api/tokens/verify", data=data) as resp:
            response = await resp.json()
            return response

    async def get_albums(self) -> AlbumsResponse:
        async with self.session.get("/api/albums") as resp:
            response = await resp.json()
            return response

    async def create_album(
        self, name: str, description: str, public: bool = True, download: bool = True
    ) -> CreateAlbumResponse:
        data = {"name": name, "description": description, "public": public, "download": download}
        async with self.session.post("/api/albums", json=data) as resp:
            response = await resp.json()
            return response

    async def upload_chunks(
        self,
        file_data: Union[BinaryIO, ProgressFileReader],
        file_name: str,
        file_uuid: str,
        file_size: int,
        session,
        server: str,
    ) -> None:
        total_chunks = (file_size + self.chunk_size - 1) // self.chunk_size
        chunk_index = 0
        dzchunkbyteoffset = 0

        # Iterates all chunks
        while chunk_index < total_chunks:
            # logger.debug(f"Processing chunk {chunk_index + 1}/{total_chunks} for {file_name}")

            chunk_data = file_data.read(self.chunk_size)
            chunk_upload_success = False
            chunk_upload_attempt = 0
            if not chunk_data:
                print("No more chunks to upload")
                break  # Exit the loop if we've reached the end of the file

            # likely using https://gitlab.com/meno/dropzone/-/wikis/faq#chunked-uploads
            # https://github.com/Dodotree/DropzonePHPchunks/issues/3
            data = aiohttp.FormData()
            data.add_field("dzuuid", file_uuid)
            data.add_field("dzchunkindex", str(chunk_index))
            data.add_field("dztotalfilesize", str(file_size))
            data.add_field("dzchunksize", str(self.chunk_size))
            data.add_field("dztotalchunkcount", str(total_chunks))
            data.add_field("dzchunkbyteoffset", str(dzchunkbyteoffset))
            data.add_field(
                "files[]",
                chunk_data,
                filename=file_name,
                content_type="application/octet-stream",
            )

            # Retries chunks if they ever fail
            while chunk_upload_attempt < self.max_chunk_retries and chunk_upload_success is False:
                try:
                    async with session.post("/api/upload", data=data) as resp:
                        response = await resp.json()
                        if response.get("success"):
                            chunk_index += 1
                            dzchunkbyteoffset += self.chunk_size
                            chunk_upload_success = True
                        else:
                            msg = f"{file_uuid} failed uploading chunk #{chunk_index}/{total_chunks} to {server} [{chunk_upload_attempt}/{self.max_chunk_retries}]"
                            logger.error(msg)
                            raise Exception(msg)
                except Exception:
                    chunk_upload_attempt += 1

            if chunk_upload_success is False:
                msg = f"Failed uploading chunks for {file_uuid} too many times times to {server}, cannot continue"
                logger.error(msg)
                raise Exception(msg)

    # TODO: This should probably move out of API
    async def upload(self, file: Path, album_id: Optional[str] = None) -> UploadResponse:
        file_size = os.stat(file).st_size
        file_mimetype = mimetypes.guess_type(file)[0] or "application/octet-stream"
        node_response = await self.get_node()
        server = "/".join(node_response["url"].split("/")[:3])

        if server not in self.server_sessions:
            logger.info(f"Using new server connection to {server}")
            self.server_sessions[server] = aiohttp.ClientSession(server, headers=self.session_headers)

        session = self.server_sessions[server]

        headers = {"albumid": album_id} if album_id else None

        file_uuid = str(uuid.uuid4())

        async with self.sem:
            retries = 0
            while retries < self.retries:
                try:
                    with open(file, "rb") as file_data:
                        with TqdmUpTo(
                            unit="B",
                            unit_scale=True,
                            unit_divisor=1024,
                            miniters=1,
                            desc=f"{file.name} [{retries + 1}/{self.retries}]",
                        ) as t:
                            with ProgressFileReader(filename=file, read_callback=t.update_to) as file_data:
                                if file_size <= self.chunk_size:
                                    chunk_data = file_data.read(self.chunk_size)
                                    data = aiohttp.FormData()
                                    data.add_field(
                                        "files[]", chunk_data, filename=file.name, content_type=file_mimetype
                                    )

                                    async with session.post("/api/upload", data=data, headers=headers) as resp:
                                        response = await resp.json()
                                        if not response.get("success"):
                                            raise Exception(f"{file.name} failed uploading without chunks")

                                        return response
                                else:
                                    logger.debug(f"{file.name} will use UUID {file_uuid}")
                                    await self.upload_chunks(
                                        file_data, file.name, file_uuid, file_size, session, server
                                    )

                                    upload_data = {
                                        "files": [
                                            {
                                                "uuid": file_uuid,
                                                "original": file.name,
                                                "type": file_mimetype,
                                                "albumid": album_id or "",
                                                "filelength": "",
                                                "age": "",
                                            }
                                        ]
                                    }
                                    finish_chunks_attempt = 0
                                    while True:
                                        try:
                                            async with session.post(
                                                "/api/upload/finishchunks", json=upload_data
                                            ) as resp:
                                                response = await resp.json()
                                                if response.get("success") is False:
                                                    msg = f"{file_uuid} failed finishing chunks to {server} [{finish_chunks_attempt + 1}/{self.max_chunk_retries}]\n{pformat(response)}"
                                                    logger.error(msg)
                                                    raise Exception(msg)
                                                # chunk_upload_success = True
                                                response["files"][0].update(
                                                    {
                                                        "fileName": file.name,
                                                        "albumid": album_id,
                                                        "filePath": str(file),
                                                        "filePathMD5": hashlib.md5(
                                                            str(file).encode("utf-8")
                                                        ).hexdigest(),
                                                        "fileNameMD5": hashlib.md5(
                                                            str(file.name).encode("utf-8")
                                                        ).hexdigest(),
                                                        "uploadSuccess": response["success"],
                                                    }
                                                )
                                                return response
                                        except Exception:
                                            finish_chunks_attempt += 1
                                            if finish_chunks_attempt >= self.max_chunk_retries:
                                                raise
                                    # TODO: Should probably return here
                except Exception:
                    logger.exception(f"Upload failed for {file.name} to {server} Attempt #{retries + 1}")
                    retries += 1
            return {"success": False, "files": [{"name": file.name, "url": ""}]}

    # TODO: This should probably move out of API
    async def upload_files(self, paths: list[Path], folder_id: Optional[str] = None) -> list[UploadResponse]:
        try:
            tasks = [self.upload(test_file, folder_id) for i, test_file in enumerate(paths)]
            responses = await tqdm_asyncio.gather(*tasks, desc="Files uploaded")
            return responses
        finally:
            # This should happen in the API client itself
            await self.session.close()
            for server_session in self.server_sessions.values():
                await server_session.close()



from io import BufferedReader
from pathlib import Path
from typing import Callable, Optional

from tqdm import tqdm


class ProgressFileReader(BufferedReader):
    def __init__(self, filename: Path, read_callback: Optional[Callable[[int, int, Optional[int]], None]] = None):
        # Don't use with because we need the file to be open for future progress
        # No idea if this causes memory issues
        f = open(filename, "rb")
        self.__read_callback = read_callback
        super().__init__(raw=f)
        self.length = Path(filename).stat().st_size

    def read(self, size: Optional[int] = None):
        calc_sz = size
        if not calc_sz:
            calc_sz = self.length - self.tell()
        if self.__read_callback:
            self.__read_callback(self.tell(), round(self.tell() / self.length), self.length)
        return super(ProgressFileReader, self).read(size)


class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        return self.update(b * bsize - self.n)  # also sets self.n = b * bsize



import argparse
import os
from pathlib import Path


def cli():
    parser = argparse.ArgumentParser(prog="bunkrr-upload", description="Bunkrr Uploader supporting parallel uploads")
    parser.add_argument("file", type=Path, help="File or directory to look for files in to upload")
    parser.add_argument(
        "-t",
        "--token",
        type=str,
        default=os.getenv("BUNKRR_TOKEN"),
        help="""API token for your account so that you can upload to a specific account/folder.
                You can also set the BUNKRR_TOKEN environment variable for this""",
    )
    parser.add_argument(
        "-f", "--folder", type=str, help="Folder to upload files to overriding the directory name if used"
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="Don't create folders or upload files",
    )
    parser.add_argument(
        "--max-chunk-size",
        action="store_true",
        help="Use the server's maximum chunk size instead of the default one",
    )
    parser.add_argument("-c", "--connections", type=int, default=2, help="Maximum parallel uploads to do at once")
    parser.add_argument(
        "--public",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Make all files uploaded public. By default they are private and not unsharable",
    )
    parser.add_argument(
        "--save",
        action=argparse.BooleanOptionalAction,
        default=True,
        help='Don\'t save uploaded file urls to a "gofile_upload_<unixtime>.csv" file',
    )
    parser.add_argument(
        "--use-config",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Whether to create and use a config file in $HOME/.config/gofile_upload/config.json",
    )
    parser.add_argument(
        "-r",
        "--retries",
        default=1,
        type=int,
        help="How many times to retry a failed upload",
    )
    parser.add_argument(
        "--chunk-retries",
        default=1,
        type=int,
        help="How many times to retry a failed chunk or chunk completion",
    )
    args = parser.parse_args()

    return args
